/*
 * kgraft_patch_bsc1088268
 *
 * Fix for CVE-2017-0861, bsc#1088268
 *
 *  Upstream commit:
 *  362bca57f5d7 ("ALSA: pcm: prevent UAF in snd_pcm_info")
 *
 *  SLE12(-SP1) commit:
 *  387619c536727b4fdfed2413dba4955eb1630044
 *
 *  SLE12-SP2 commit:
 *  206b25525b21a590445aa55a1be79fe38805e485 ("stable 4.4.106")
 *
 *  SLE12-SP3 commit:
 *  7e856710dfdf632bd860841fa444d89b6607bb7b ("stable 4.4.106")
 *
 *  Copyright (c) 2018 SUSE
 *  Author: Nicolai Stange <nstange@suse.de>
 *
 *  Based on the original Linux kernel code. Other copyrights apply.
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, see <http://www.gnu.org/licenses/>.
 */

#include <linux/kernel.h>
#include <linux/module.h>
#include <linux/kallsyms.h>
#include <sound/core.h>
#include <sound/control.h>
#include <sound/pcm.h>
#include "kgr_patch_bsc1088268.h"


#if IS_ENABLED(CONFIG_SND_PCM)

#if !IS_MODULE(CONFIG_SND_PCM)
#error "KGR patch supports only CONFIG_SND_PCM=m."
#endif

#define KGR_PATCHED_MODULE "snd_pcm"


struct list_head *kgr_snd_pcm_devices;
struct mutex *kgr_register_mutex;

static int (*kgr_snd_pcm_info_user)(struct snd_pcm_substream *substream,
				    struct snd_pcm_info __user * _info);

static struct {
	char *name;
	void **addr;
} kgr_funcs[] = {
	{ "snd_pcm:snd_pcm_devices", (void *)&kgr_snd_pcm_devices },
	{ "snd_pcm:register_mutex", (void *)&kgr_register_mutex },
	{ "snd_pcm:snd_pcm_info_user", (void *)&kgr_snd_pcm_info_user },
};


/* inlined */
static struct snd_pcm *kgr_snd_pcm_get(struct snd_card *card, int device)
{
	struct snd_pcm *pcm;

	list_for_each_entry(pcm, kgr_snd_pcm_devices, list) {
		if (pcm->card == card && pcm->device == device)
			return pcm;
	}
	return NULL;
}

/* inlined */
static int kgr_snd_pcm_next(struct snd_card *card, int device)
{
	struct snd_pcm *pcm;

	list_for_each_entry(pcm, kgr_snd_pcm_devices, list) {
		if (pcm->card == card && pcm->device > device)
			return pcm->device;
		else if (pcm->card->number > card->number)
			return -1;
	}
	return -1;
}



/* patched */
int kgr_snd_pcm_control_ioctl(struct snd_card *card,
			      struct snd_ctl_file *control,
			      unsigned int cmd, unsigned long arg)
{
	switch (cmd) {
	case SNDRV_CTL_IOCTL_PCM_NEXT_DEVICE:
		{
			int device;

			if (get_user(device, (int __user *)arg))
				return -EFAULT;
			mutex_lock(kgr_register_mutex);
			device = kgr_snd_pcm_next(card, device);
			mutex_unlock(kgr_register_mutex);
			if (put_user(device, (int __user *)arg))
				return -EFAULT;
			return 0;
		}
	case SNDRV_CTL_IOCTL_PCM_INFO:
		{
			struct snd_pcm_info __user *info;
			unsigned int device, subdevice;
			int stream;
			struct snd_pcm *pcm;
			struct snd_pcm_str *pstr;
			struct snd_pcm_substream *substream;
			int err;

			info = (struct snd_pcm_info __user *)arg;
			if (get_user(device, &info->device))
				return -EFAULT;
			if (get_user(stream, &info->stream))
				return -EFAULT;
			if (stream < 0 || stream > 1)
				return -EINVAL;
			if (get_user(subdevice, &info->subdevice))
				return -EFAULT;
			mutex_lock(kgr_register_mutex);
			pcm = kgr_snd_pcm_get(card, device);
			if (pcm == NULL) {
				err = -ENXIO;
				goto _error;
			}
			pstr = &pcm->streams[stream];
			if (pstr->substream_count == 0) {
				err = -ENOENT;
				goto _error;
			}
			if (subdevice >= pstr->substream_count) {
				err = -ENXIO;
				goto _error;
			}
			for (substream = pstr->substream; substream;
			     substream = substream->next)
				if (substream->number == (int)subdevice)
					break;
			if (substream == NULL) {
				err = -ENXIO;
				goto _error;
			}
			/*
			 * Fix CVE-2017-0861
			 *  +1 line
			 */
			mutex_lock(&pcm->open_mutex);
			err = kgr_snd_pcm_info_user(substream, info);
			/*
			 * Fix CVE-2017-0861
			 *  +1 line
			 */
			mutex_unlock(&pcm->open_mutex);
		_error:
			mutex_unlock(kgr_register_mutex);
			return err;
		}
	case SNDRV_CTL_IOCTL_PCM_PREFER_SUBDEVICE:
		{
			int val;

			if (get_user(val, (int __user *)arg))
				return -EFAULT;
			control->preferred_subdevice[SND_CTL_SUBDEV_PCM] = val;
			return 0;
		}
	}
	return -ENOIOCTLCMD;
}



static int kgr_patch_bsc1088268_kallsyms(void)
{
	unsigned long addr;
	int i;

	for (i = 0; i < ARRAY_SIZE(kgr_funcs); i++) {
		/* mod_find_symname would be nice, but it is not exported */
		addr = kallsyms_lookup_name(kgr_funcs[i].name);
		if (!addr) {
			pr_err("kgraft-patch: symbol %s not resolved\n",
				kgr_funcs[i].name);
			return -ENOENT;
		}

		*(kgr_funcs[i].addr) = (void *)addr;
	}

	return 0;
}

static int kgr_patch_bsc1088268_module_notify(struct notifier_block *nb,
					      unsigned long action, void *data)
{
	struct module *mod = data;
	int ret;

	if (action != MODULE_STATE_COMING || strcmp(mod->name, KGR_PATCHED_MODULE))
		return 0;

	ret = kgr_patch_bsc1088268_kallsyms();
	WARN(ret, "kgraft-patch: delayed kallsyms lookup failed. System is broken and can crash.\n");

	return ret;
  }

 static struct notifier_block kgr_patch_bsc1088268_module_nb = {
	.notifier_call = kgr_patch_bsc1088268_module_notify,
	.priority = INT_MIN+1,
 };

int kgr_patch_bsc1088268_init(void)
{
	int ret;

	mutex_lock(&module_mutex);
	if (find_module(KGR_PATCHED_MODULE)) {
		ret = kgr_patch_bsc1088268_kallsyms();
		if (ret)
			goto out;
	}

	ret = register_module_notifier(&kgr_patch_bsc1088268_module_nb);
out:
	mutex_unlock(&module_mutex);
	return ret;
}

void kgr_patch_bsc1088268_cleanup(void)
{
	unregister_module_notifier(&kgr_patch_bsc1088268_module_nb);
}

#endif /* IS_ENABLED(CONFIG_SND_PCM) */
